package org.oasisopen.sca;

import java.io.BufferedReader;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;

import org.oasisopen.sca.spi.SCAClientFactorySPI;

/**
 * The SCAClientFactory can be used by non-SCA managed code to lookup services
 * that exist in a SCADomain.
 * 
 * @author OASIS Open
 */
public class SCAClientFactory {

    /**
     * This is the name of the System Property that we look at to determine the SPI implementation 
     * to use for the SCAClientFactory.
     */
    private static final String SCA_CLIENT_FACTORY_PROVIDER_KEY = SCAClientFactory.class.getName();

    /**
     * This is the name of the file that we load from the ClassPath to determine the SPI implementation 
     * to use for the SCAClientFactory.
     */
    private static final String SCA_CLIENT_FACTORY_PROVIDER_META_INF_SERVICE 
        = "META-INF/services/" + SCA_CLIENT_FACTORY_PROVIDER_KEY;

    /**
     * The default implementation of the SCAClientFactory.
     */
    private static final SCAClientFactory defaultFactory = new SCAClientFactory(); 
    
    /**
     * Reference to the SCAClientFactorySPI implementation that will be used by this SCAClientFactory
     * instance.
     */
    private final SCAClientFactorySPI providerFactory;
    
    /**
     * Reference to the ClassLoader that will be used by this SCAClientFactory instance.
     */
    private final ClassLoader cl;
    
    /**
     * Private Constructor that will initialise the SCAClientFactorySPI implementation.
     * 
     * @throws SCARuntimeException Failed to instantiate the SCAClientFactory 
     */
    private SCAClientFactory() {
        cl = getThreadContextClassLoader();
        providerFactory = createSCAClientFactorySPIImplentation();
    }
    
    /**
     * Creates a new instance of the SCAClient that can be used to lookup SCA Services
     * 
     * @return A new SCAClient 
     */
    public static SCAClient createSCAClient() {
        final SCAClient scaClient = defaultFactory.createSCAClientUsingProvider();
        return scaClient;
    }

    /**
     * Creates a new SCAClient using the providerFactory.
     * 
     * @return A new SCAClient.
     */
    private SCAClient createSCAClientUsingProvider()
    {
        final SCAClient scaClient = providerFactory.createSCAClient();
        return scaClient;
    }

    /**
     * Creates an instance of the SCAClientFactorySPI implementation. This will discover the
     * SCAClientFactorySPI Implementation and instantiate the providers implementation.
     * 
     * @return new instance of the SCAClientFactorySPI
     * @throws SCARuntimeException Failed to create SCAClientFactorySPI Implementation.
     */
    private SCAClientFactorySPI createSCAClientFactorySPIImplentation()
    {
        final String factoryImplClassName = discoverProviderFactoryImplClass();
        final Class<? extends SCAClientFactorySPI> factoryImplClass 
            = loadProviderFactoryClass(factoryImplClassName);
        final SCAClientFactorySPI factory = instantiateProviderFactoryClass(factoryImplClass);
        return factory;
    }

    /**
     * Gets the Context ClassLoader for the current Thread.
     * 
     * @return The Context ClassLoader for the current Thread.
     */
    private ClassLoader getThreadContextClassLoader() {
        final ClassLoader threadClassLoader = Thread.currentThread().getContextClassLoader();
        return threadClassLoader;
    }

    /**
     * Attempts to discover the class name for the SCAClientFactorySPI implementation from the
     * System Properties or the classpath.
     * 
     * @return The class name of the SCAClientFactorySPI implementation
     * @throw SCARuntimeException Failed to find implementation for SCAClientFactorySPI.
     */
    private String discoverProviderFactoryImplClass() throws SCARuntimeException{
        String providerClassName = checkSystemPropertiesForSPIClassName();
        if (providerClassName != null) {
            return providerClassName;
        }
        
        providerClassName = checkMETAINFServicesForSIPClassName();
        if (providerClassName != null) {
            return providerClassName;
        }
        
        throw new SCARuntimeException("Failed to find implementation for SCAClientFactory");
    }
    
    /**
     * Attempts to find the class name for the SCAClientFactorySPI implementation from the 
     * System Properties.
     * 
     * @return The class name for the SCAClientFactorySPI implementation or <code>null</code>
     * if not found.
     */
    private String checkSystemPropertiesForSPIClassName()
    {
        final String providerClassName = System.getProperty(SCA_CLIENT_FACTORY_PROVIDER_KEY);
        if (providerClassName != null && providerClassName.length() > 0) {
            return providerClassName;
        }
        
        return null;
    }

    /**
     * Attempts to find the class name for the SCAClientFactorySPI implementation from the 
     * META-INF/services.
     * 
     * @return The class name for the SCAClientFactorySPI implementation or <code>null</code>
     * if not found.
     */
    private String checkMETAINFServicesForSIPClassName() {
        final URL url = cl.getResource(SCA_CLIENT_FACTORY_PROVIDER_META_INF_SERVICE);
        if (url == null) {
            return null;
        }
        
        InputStream in = null; 
        try {
            in = url.openStream();
            BufferedReader reader = null;
            try {
                reader = new BufferedReader(new InputStreamReader(in, "UTF-8"));

                String line;
                while ((line = readNextLine(reader)) != null) {
                    if (!line.startsWith("#") && line.length() > 0) {
                        return line;
                    }
                }
                
                return null;
            } finally {
                closeStream(reader);
            }
        } catch (IOException ex) {
            throw new SCARuntimeException("Failed to discover SCAClientFactory provider", ex);
        } finally {
            closeStream(in);
        }
    }

    /**
     * Reads the next line from the reader and returns the trimmed version of that line/
     * 
     * @param reader The reader from which to read the next line
     * @return The trimmed next line or <code>null</code> if the end of the steam has been reached
     * @throws IOException I/O error occurred while reading from Reader
     */
    private String readNextLine(BufferedReader reader) throws IOException {
        String line = reader.readLine();
        if (line != null) {
            line = line.trim();
        }
        return line;
    }
    
    /**
     * Loads the specified SCAClientFactorySPI Implementation class.
     * 
     * @param factoryImplClassName The name of the SCAClientFactorySPI Implementation class to load
     * @return The specified SCAClientFactorySPI Implementation class
     * @throws SCARuntimeException Failed to load the SCAClientFactorySPI Implementation class 
     */
    private Class<? extends SCAClientFactorySPI> loadProviderFactoryClass(String factoryImplClassName) 
        throws SCARuntimeException {
        
        try {
            final Class<?> provderClass = cl.loadClass(factoryImplClassName);
            final Class<? extends SCAClientFactorySPI> providerFactoryClass 
                = (Class<? extends SCAClientFactorySPI>) provderClass;
            return providerFactoryClass;
        } catch (ClassNotFoundException ex) {
            throw new SCARuntimeException("Failed to load provider factory class " + factoryImplClassName, ex);
        }
    }

    /**
     * Instantiate an instance of the specified SCAClientFactorySPI Implementation class.
     *  
     * @param factoryImplClass The SCAClientFactorySPI Implementation class to instantiate.
     * @return An instance of the SCAClientFactorySPI Implementation class
     * @throws SCARuntimeException Failed to instantiate the specified specified SCAClientFactorySPI Implementation class  
     */
    private SCAClientFactorySPI instantiateProviderFactoryClass(Class<? extends SCAClientFactorySPI> factoryImplClass)
        throws SCARuntimeException {
        
        try {
            final SCAClientFactorySPI provider = factoryImplClass.newInstance();
            return provider;
        } catch (SCARuntimeException ex) {
            throw ex;
        } catch (Throwable ex) {
            throw new SCARuntimeException("Failed to instantiate SCAClient Provider Factory class " + factoryImplClass, ex);
        }
    }
    
    /**
     * Utility method for closing Closeable Object.
     * 
     * @param closeable The Object to close.
     */
    private void closeStream(Closeable closeable) {
        if (closeable != null) {
            try{
                closeable.close();
            } catch (IOException ex) {
                throw new SCARuntimeException("Failed to close stream", ex);
            }
        }
    }
}
